/*
    Extent File System - Volume data file storage
    Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	convert.cpp

	This is the restriper. It takes an XML extent file descriptor and
	a series of plane bitmaps as input, and generates an extent file
	as output.
*/

#define DECLARE_CONVERT

#include <stdio.h>
extern "C" 
{
#include <jpeglib.h>
}

#include <windows.h>
#include "../ExtentFile/ExtentFile.h"
#include <zlib.h>

using namespace extfile;

//! Structure for describing a stripe
struct StripeData
{
	//! File handle for the stripe
	FileWrapper data;
	//! Current size of stripe (in blocks)
	Int32 dataSize;
};

void extfile::ExtentFileDescriptor::convert(StripeFunc stripefunc)
{
	int i,j;


	// Create all stripes
	StripeData *sd=(StripeData*)malloc(sizeof(StripeData)*stripeCount);
	for(i=0;i<stripeCount;i++)
	{
		char fname[1024];
		sprintf(fname,"%s/%s_",(const char*)stripes[i].path,(const char*)stripes[i].data);
		if(!sd[i].data.open(fname,FileWrapper::WRITE))
		{
			printf("ERROR: Could not open %s\n",fname);
			printf("Please make sure that the path in the XML descriptor is correct.\n");
			return;
		}
		sd[i].dataSize=0;
	}

	// Allocate address table
	AddressEntry *addressTable;
	Int32 addressTableSize;

	addressTableSize=extOff[resolutions];
	addressTable=(AddressEntry*)malloc(addressTableSize*sizeof(AddressEntry));

	Int32 tWriteCount=0;

	// Allocate output extent memory (complete slices of extents)
	UInt8 *outCache[MAXZOOM];
	for(i=0;i<resolutions;i++)
	{
		outCache[i]=(UInt8*)malloc(extSize*extX[i]*extY[i]);
		memset(outCache[i],0,extSize*extX[i]*extY[i]);
	}
	// Allocate temporary output data buffer
	UInt8 *outData=(UInt8*)malloc(1048576);
	// Allocate source slice buffers
	UInt8 ***sourceSlices=(UInt8***)malloc(sizeof(UInt8**)*channels);
	for(i=0;i<channels;i++)
	{
		sourceSlices[i]=(UInt8**)malloc(sizeof(UInt8*)*fullExtentSize.z);
		for(j=0;j<fullExtentSize.z;j++)
			sourceSlices[i][j]=(UInt8*)malloc(sources[i].width*sources[i].height*channelSize(sources[i].channel));
	}


	// Traverse all slices of extents
	for(int z=0;z<extZ[0];z++)
	{
		printf("Creating extents: %d%%    \r",z*100/extZ[0]);
		// Load slices
		for(i=0;i<channels;i++)
		{
			for(j=0;j<fullExtentSize.z;j++)
			{
				// If slice should exist, try to load it
				if(j+z*extentSize.z<sources[i].count)
				{
					char fname[1024];
					sprintf(fname,(const char *)sources[i].name,sources[i].offset+j+z*extentSize.z);
					// Load raw format slice
					if(!strcmp(sources[i].format,"raw"))
					{
						FILE *fptr=fopen(fname,"rb");
						if(fptr==NULL)
						{
							printf("ERROR: Could not open source file %s\n",fname);
							printf("Please make sure that the path in the XML descriptor is correct.\n");
							return;
						}
						fread(sourceSlices[i][j],sources[i].width*sources[i].height,channelSize(sources[i].channel),fptr);
						fclose(fptr);
					}
					// Load JPEG format slice
					else if(!strcmp(sources[i].format,"jpg"))
					{
						FILE *fptr=fopen(fname,"rb");
						if(fptr==NULL)
						{
							printf("ERROR: Could not open source file %s\n",fname);
							printf("Please make sure that the path in the XML descriptor is correct.\n");
							return;
						}
						struct jpeg_decompress_struct cinfo;
						struct jpeg_error_mgr jerr;

						cinfo.err = jpeg_std_error(&jerr);
						jpeg_create_decompress(&cinfo);

						jpeg_stdio_src(&cinfo, fptr);

						jpeg_read_header(&cinfo, TRUE);
						jpeg_start_decompress(&cinfo);
						// Verify that slice corresponds to indications in XML descriptor
						if((Int32)cinfo.output_width!=sources[i].width || (Int32)cinfo.output_height!=sources[i].height || ExtentFileDescriptor::R8G8B8!=sources[i].channel)
						{
							printf("ERROR: Invalid jpeg image, found %dx%d/%d, expected %dx%d/%d",cinfo.output_width,cinfo.output_height,ExtentFileDescriptor::R8G8B8,sources[i].width,sources[i].height,sources[i].channel);
							return;
						}
						while (cinfo.output_scanline < cinfo.output_height) 
						{
							UInt8 *buffer=&((UInt8*)sourceSlices[i][j])[cinfo.output_scanline*cinfo.output_width * cinfo.output_components];
							jpeg_read_scanlines(&cinfo, &buffer, 1);
						}

						jpeg_finish_decompress(&cinfo);
						jpeg_destroy_decompress(&cinfo);
						fclose(fptr);
					}
					else if(!strcmp(sources[i].format,"otherformat"))
					{
						//
						// Here is a good place to insert handlers for more image formats...
						//
						printf("ERROR: Unsupported format %s\n",(const char *)sources[i].format);
						return;
					}
					// Print error message if format is unknown
					else
					{
						printf("ERROR: Unsupported format %s\n",(const char *)sources[i].format);
						return;
					}
				}
				else
					memset(sourceSlices[i][j],0,sources[i].width*sources[i].height*channelSize(sources[i].channel));
			}
		}

		// Reorganize the data into extents
		for(i=0;i<channels;i++)
		{
			int loff=0;
			int sls=channelSize(sources[i].channel);
			int dls=channelSize(channelFormat[i]);
			for(j=0;j<i;j++)
				loff+=fullExtentSize.x*fullExtentSize.y*fullExtentSize.z*channelSize(channelFormat[j]);

			// Traverse all resolutions
			for(int r=0;r<resolutions;r++)
			{
				// Traverse all voxels
				for(int y=0;y<extY[r];y++)
				{
					for(int x=0;x<extX[r];x++)
					{
						int oe=(x+y*extX[r])*extSize;
						UInt8 sl[8];
						memset(sl,0,8);

						for(j=0;j<fullExtentSize.z;j++)
						{
							int rsn=j+z*extentSize.z;
							if((rsn&((1<<r)-1))!=0)
								continue;
							int tj=(rsn>>r)%extentSize.z;
							if(tj==0 && j==fullExtentSize.z-1)
								tj=fullExtentSize.z-1;

							int py=y*extentSize.y*(1<<r);

							for(int k=0;k<fullExtentSize.y;k++)
							{
								int px=x*extentSize.x*(1<<r);
								int o0=(px+py*sources[i].width)*sls;
								int o1=oe+loff+(k*fullExtentSize.x+tj*fullExtentSize.x*fullExtentSize.y)*dls;
								for(int l=0;l<fullExtentSize.x;l++)
								{
									if(px<sources[i].width && py<sources[i].height)
										memcpy(sl,&sourceSlices[i][j][o0],sls);
									else
										memset(sl,0,sls);

									// Convert colors
									if(channelFormat[i]==L16 && sources[i].channel==L16_BIG)
									{
										UInt8 t=sl[0];
										sl[0]=sl[1];
										sl[1]=t;
									}
									if(channelFormat[i]==B8G8R8X8 && sources[i].channel==R8G8B8)
									{
										UInt8 t=sl[0];
										sl[0]=sl[2];
										sl[2]=t;
										sl[3]=0;
									}

									//
									// Here is a good place to add more color space conversions
									//
									memcpy(&outCache[r][o1],sl,dls);
									o0+=(1<<r)*sls;
									o1+=dls;
								}
								py+=1<<r;
							}
						}
					}
				}
			}
		}

		// Write the completed extents to disk
		for(int r=0;r<resolutions;r++)
		{
			if(((z+1)&((1<<r)-1))!=0 && z<extZ[0]-1)
				continue;

			// This extent is now complete, write it
			for(i=0;i<extX[r]*extY[r];i++)
			{
				int nzp=0;
				for(j=0;j<extSize;j++)
				{
					if(outCache[r][j+i*extSize])
					{
						nzp++;
						break;
					}
				}
				int eid=i+(z>>r)*extX[r]*extY[r]+extOff[r];

				int file=stripefunc(eid,stripeCount);

				int offset=-1;
				int length=-1;
				int invalid=-2;
				if(nzp)
				{
					// This extent contains something
					unsigned long outSize;
					
					if(flags&COMPRESSED)
					{
						// Compress extent
						outSize=1048576;
						compress(outData,&outSize,&outCache[r][i*extSize],extSize);
					}
					else
					{
						// Just copy it
						memcpy(outData,&outCache[r][i*extSize],extSize);
						outSize=extSize;
					}
					memset(&outData[outSize],0,(outSize+blockSize-1)/blockSize*blockSize-outSize);
					length=outSize;
					outSize=(outSize+blockSize-1)/blockSize;
					offset=sd[file].dataSize;



					sd[file].data.write(outData,outSize*blockSize);
					sd[file].dataSize+=outSize;
				}
				addressTable[eid].offset=offset;
				addressTable[eid].length=length;
			}
			memset(outCache[r],0,extSize*extX[r]*extY[r]);
		}
	}

	printf("Creating extents: Done    \n");
	printf("Reorganizing files\n");

	// The stripes have been written, now we reorder them in the final files and 
	// split them. This is required since the extents have been written in a fairly
	// random sequence, and we want them ordered by resolution, lowest resolutions
	// first.
	for(i=0;i<stripeCount;i++)
	{
		char fname[1024];
		sprintf(fname,"%s/%s_",(const char*)stripes[i].path,(const char*)stripes[i].data);
		sd[i].data.close();
		if(!sd[i].data.open(fname,FileWrapper::READ))
		{
			printf("ERROR: Could not open %s\n",fname);
			return;
		}
		FileWrapper fof;
		sprintf(fname,"%s/%s",(const char*)stripes[i].path,(const char*)stripes[i].data);
		if(!fof.open(fname,FileWrapper::WRITE))
		{
			printf("ERROR: Could not open %s\n",fname);
			return;
		}
		Int32 ec=extOff[resolutions];
		Int32 off=0;
		for(j=0;j<ec;j++)
		{
			if((int)stripefunc(j,stripeCount)==i && addressTable[j].length!=0xffffffff)
			{
				Int32 l=(addressTable[j].length+blockSize-1)/blockSize;
				sd[i].data.seek((Int64)addressTable[j].offset*blockSize);
				sd[i].data.read(outData,l*blockSize);
				fof.write(outData,l*blockSize);
				addressTable[j].offset=off;
				off+=l;
			}
		}
		fof.close();
		sd[i].data.close(true);
	}

	// And to finish it off we write out the index files
	for(i=0;i<stripeCount;i++)
	{
		char fname[1024];
		sprintf(fname,"%s/%s",(const char*)stripes[i].path,(const char*)stripes[i].index);
		FILE *ndx=fopen(fname,"wb");
		if(ndx==NULL)
		{
			printf("ERROR: Could not open %s\n",fname);
			return;
		}
		fwrite(addressTable,addressTableSize,sizeof(AddressEntry),ndx);
		fclose(ndx);
		sd[i].data.close();
	}

	// Free allocated memory
	free(addressTable);
	for(i=0;i<resolutions;i++)
		free(outCache[i]);
	free(outData);
	for(i=0;i<channels;i++)
	{
		for(j=0;j<fullExtentSize.z;j++)
			free(sourceSlices[i][j]);
		free(sourceSlices[i]);
	}
	free(sourceSlices);

}

//! Main function
void main(int argc,	char *argv[])
{
	if(argc==1)
	{
		printf("Usage: convert <xmlfile1> [xmlfile2] ...\n");
		return;
	}
	for(int i=1;i<argc;i++)	
	{
		ExtentFileDescriptor efd;

		if(!efd.loadFromXML(argv[i]))
			fprintf(stderr,"ERROR: Error parsing file '%s'\n",argv[i]);

		efd.convert(extfile::defaultStripe);
	}

}