/*
    Extent File System - Volume data file storage
    Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	CachedExtentFile.cpp

	This is the implementation of the CachedExtentFile class
*/

#include "ExtentFile.h"
#include <math.h>

namespace extfile
{

	bool CachedExtentFile::open(const char *configFileName, const UInt32 cs)
	{
		int i,j,k;
		if(active)
			close();

		// Open the XML configuration file
		ExtentFile::open(configFileName);

		// Calculate size of cache in extents
		cacheSize=cs/extSize;
		baseSize[0]=0;
		for(i=1;i<efd.channels;i++)
			baseSize[i]=baseSize[i-1]+extSizeX*extSizeY*extSizeZ*efd.channelSize(efd.channelFormat[i-1]);

		// Allocate cache mangement structures
		lruListItem=(LoadedExtent *)malloc(extCount*sizeof(LoadedExtent));
		cacheBase=(UInt8 *)malloc(cacheSize*extSize);
		cacheHead=cacheTail=NULL;
		bottom=(UInt8 *)malloc(extX[zoomLevels-1]*extY[zoomLevels-1]*extZ[zoomLevels-1]*extSize);
		zeroBuffer=(UInt32 *)malloc(extSize);
		memset(zeroBuffer,0x00,extSize);

		memset(lruListItem,0,extCount*sizeof(LoadedExtent));
		unusedCacheSlots=cacheSize;

		// Load bottom extent
		for(k=0;k<extZ[zoomLevels-1];k++)
		{
			for(j=0;j<extY[zoomLevels-1];j++)
			{
				for(i=0;i<extX[zoomLevels-1];i++)
				{
					int en=extOff[zoomLevels-1]+i+j*extX[zoomLevels-1]+k*extX[zoomLevels-1]*extY[zoomLevels-1];
					lruListItem[en].extent=en;
					lruListItem[en].address=NULL;
					lruListItem[en].finalAddress=(UInt32 *)&bottom[extSize*(i+j*extX[zoomLevels-1]+k*extX[zoomLevels-1]*extY[zoomLevels-1])];

					if(address[en<<1]==0xffffffff)
					{
						memset(lruListItem[en].finalAddress,0,extSize);
					}
					else
					{
						ExtentFile::loadExtent(en,lruListItem[en].finalAddress);
					}
				}
			}
		}

		// Set the addresses of all black extents
		for(i=4+4*zoomLevels;i<extCount;i++)
		{
			if(address[i<<1]==0xffffffff)
			{
				lruListItem[i].finalAddress=zeroBuffer;
				lruListItem[i].address=NULL;
			}
		}

		active=true;
		return true;
	}

	bool CachedExtentFile::close()
	{
		free(lruListItem);
		free(cacheBase);
		free(bottom);
		free(zeroBuffer);

		active=false;
		return true;
	}


	bool CachedExtentFile::loadExtent(int en)
	{
		// Verify that requested extent is in range
		if(en>=extCount || en<0)
			return false;

		if(lruListItem[en].finalAddress)
		{
			// The extent is already in our cache
			if(lruListItem[en].address)
			{
				// Just put it at the head of the list
				if(lruListItem[en].prev)
					lruListItem[en].prev->next=lruListItem[en].next;
				else
					cacheHead=lruListItem[en].next;
				if(lruListItem[en].next)
					lruListItem[en].next->prev=lruListItem[en].prev;
				else
					cacheTail=lruListItem[en].prev;
				lruListItem[en].next=cacheHead;
				lruListItem[en].prev=NULL;
				if(cacheHead)
					cacheHead->prev=&lruListItem[en];
				else
					cacheTail=&lruListItem[en];
				cacheHead=&lruListItem[en];

			}
			return false;
		}
		else
		{
			// It is in no list, add it to the items that must be loaded.
			if(unusedCacheSlots)
			{
				// There is still an available cache slot
				lruListItem[en].slot=cacheSize-unusedCacheSlots;
				unusedCacheSlots--;
			}
			else
			{
				// There are no more cache slots, something must be dumped
				LoadedExtent *last=cacheTail;
				int cs=cacheTail->slot;
				cacheTail->prev->next=NULL;
				cacheTail=cacheTail->prev;
				last->prev=NULL;
				last->next=NULL;
				last->finalAddress=NULL;
				last->address=NULL;
				lruListItem[en].slot=cs;
			}
			// Fill out extent descriptor
			lruListItem[en].extent=en;
			lruListItem[en].address=(UInt32 *)&cacheBase[extSize*lruListItem[en].slot];
			lruListItem[en].next=NULL;
			lruListItem[en].prev=NULL;

			// Add to list of extents in cache
			LoadedExtent *le=&lruListItem[en];
			le->next=cacheHead;
			le->prev=NULL;
			if(cacheHead)
				cacheHead->prev=le;
			else
				cacheTail=le;
			cacheHead=le;

			// Load physical extent
			ExtentFile::loadExtent(le->extent,le->address);
			le->finalAddress=le->address;
		}
		return true;
	}

	void CachedExtentFile::getSlice(float xp, float yp, float zp, float ux, float uy, float uz, float vx, float vy, float vz, float zoom, int w, int h, void *data, int c)
	{
		int x, y;
		int cx, cy, cz, ocx, ocy, ocz, dux, duy, duz, dvx, dvy, dvz;

		// Evaluate resolution required for this slice
		int datasetZoom=0;
		float tz=1/zoom;
		int zf=1;

		while(tz>=2.0f)
		{
			tz/=2.0;
			datasetZoom++;
			zf*=2;
		}

		// Scaling factor (adjust for resolution and requested zoom)
		float sf=1/zf/zoom;

		// Length of direction vectors (for normalization)
		float lu=(float)sqrt(ux*ux+uy*uy+uz*uz);
		float lv=(float)sqrt(vx*vx+vy*vy+vz*vz);

		// Reevaluate direction vectors taking into account aspect ratio and scale, and 
		// convert to 16.16 fixed point.
		dux=(int)(ux/lu*65536*sf/voxelSizeX);
		duy=(int)(uy/lu*65536*sf/voxelSizeY);
		duz=(int)(uz/lu*65536*sf/voxelSizeZ);
		dvx=(int)(vx/lv*65536*sf/voxelSizeX);
		dvy=(int)(vy/lv*65536*sf/voxelSizeY);
		dvz=(int)(vz/lv*65536*sf/voxelSizeZ);
		
		// Same for slice position, calculate top left corner coordinate in fixed point
		ocx=(int)((xp-w*ux/zoom/2-h*vx/zoom/2)*65536/zf/voxelSizeX);
		ocy=(int)((yp-w*uy/zoom/2-h*vy/zoom/2)*65536/zf/voxelSizeY);
		ocz=(int)((zp-w*uz/zoom/2-h*vz/zoom/2)*65536/zf/voxelSizeZ);

		// Fixed point walkthrough in destination space
		UInt8 *op=(UInt8*)data;
		int cs=efd.channelSize(efd.channelFormat[c]);
		for(y=0;y<h;y++)
		{
			cx=ocx;
			cy=ocy;
			cz=ocz;
			
			for(x=0;x<w;x++)
			{
				// Get the voxel from the dataset (point sampling)
				*(UInt32*)op=getVoxel(datasetZoom,cx>>16,cy>>16,cz>>16,c);
				op+=cs;
				
				cx+=dux;
				cy+=duy;
				cz+=duz;
			}

			ocx+=dvx;
			ocy+=dvy;
			ocz+=dvz;
		}
	}

}

