/*
    Extent File System - Volume data file storage
    Copyright (C) 2000-2002 Sebastian Gerlach EPFL-LSP

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	ExtentFile.cpp

	This is the implementation of the ExtentFile class
*/

/*! 
\mainpage Extent File System Reference

\section intro Introduction
<p>Large 3D volumes need to be segmented into smaller size subvolumes both for ensuring 
efficient access to the data and for storing the data on multiple disks. This is 
especially useful for large medical 3D volume data sets such as CT, MRI, or cryosections.

We propose a
striped file library which distributes the subvolumes onto multiple disks (multiple 
subfiles). The dataset is split into regular cubic subvolumes called extents, which are
distributed onto the available subfiles.
<p>To further improve data access speed, the data is stored at multiple resolutions in the 
subfiles, so that extents with a resolution near to the target display resolution can 
be read. 
<p>The data contained in the subfiles requires attached metadata, such as its physical 
resolution, image size, and component files. This metadata is stored in a standard XML 
file for optimal readability and ease of maintenance. Subfiles and 
associated metadata are denominated Extent Files.
<p>The library is provided under the GPL. Currently, the source code is provided with project
files for Windows platforms. This library is used for the data storage on our 
<a href="http://visiblehuman.epfl.ch">Visible Human Server</a>.
\section comp Library components
<p>The library is composed of three parts:
<ul>
<li>The extent file library
<li>The extent file conversion program
<li>A sample program showing how to use the two provided interfaces
</ul>
<p>The extent file library provides two class for using extent files:
<ul><li>ExtentFile: this is the standard class, providing low-level access primitives
such as reading a single extent, or an arbitrary region from the dataset.
<li>CachedExtentFile: this class uses a memory cache for loaded extentsl, and provides
finer grain (extent independent) access to the data, such as reading individual voxels
and extracting arbitrary slices from the dataset.</ul>
<p>The conversion program takes a descriptor file as well as a collection of planar
bitmaps as input, and reorganizes them to an extent file as specified in the descriptor.
\section build Building the library
<p>To build the library and sample application, you will require libxml2
(<a href="http://www.xmlsoft.org/">www.xmlsoft.org</a>), zlib
(<a href="http://www.gzip.org/zlib/">www.gzip.org/zlib</a>), and libjpeg
(<a href="http://www.ijg.org/">www.ijg.org</a>). They are
required for reading the XML file descriptors, creating compressed
extent files, and reading jpeg source images respectively.
*/

#include "ExtentFile.h"
#include <zlib.h>

namespace extfile
{
	//! Default striping function
	/*!
	This function simply returns extent number modulo number of stripes.
	\param in extent number
	\param sc number of stripes
	*/
	UInt32 defaultStripe(UInt32 in, UInt32 sc)
	{
		return in%sc;
	}

	//! Separates string into comma separated parts
	/*!
	As opposed to strtok, this will return zero length strings when two successive
	separators are found .
	*/
	static char *strcsv(char *src)
	{
		static char *cp;
		if(src)
			cp=src;
		char *rv=cp;
		int qc=0;
		for(char *a=cp;(*a!=',' || (qc&1)) && *a!=0;a++)
		{
			if(*a=='"')
				qc++;
		}
		*a=0;
		cp=a+1;
		if(*rv=='"')
		{
			a[-1]=0;
			return rv+1;
		}
		return rv;
	}

	bool ExtentFile::loadLabels(const char *fname)
	{
		// Open file
		FILE *fptr=fopen(fname,"rb");
		char buf[1024];

		if(fptr==NULL)
			return false;

		// Count number of labels in file
		int maxndx=0;
		while(fgets(buf,1024,fptr))
		{
			strtok(buf,"\r\n");
			char *ndx=strcsv(buf);
			int i=atoi(ndx);
			if(i>maxndx)
				maxndx=i;
		}

		// Allocate label structure
		maxndx++;
		labels=(LabelDescription*)malloc(maxndx*sizeof(LabelDescription));
		labelCount=maxndx;

		for(int i=0;i<maxndx;i++)
		{
			labels[i].display=false;
			labels[i].use=false;
		}

		// Load labels from file
		fseek(fptr,0,0);
		while(fgets(buf,1024,fptr))
		{
			strtok(buf,"\r\n");
			char *ndx=strcsv(buf);
			char *use=strcsv(NULL);
			char *name=strcsv(NULL);
			char *side=strcsv(NULL);
			char *loc=strcsv(NULL);
			char *sys=strcsv(NULL);
			int i=atoi(ndx);
			labels[i].use=(atoi(use)!=0);
			strcpy(labels[i].name,name);
			strcpy(labels[i].side,side);
			strcpy(labels[i].location,loc);
			strcpy(labels[i].system,sys);
			labels[i].display=labels[i].use;
		}
		fclose(fptr);

		labels[0].display=false;

		return true;
	}


	void ExtentFile::loadExtent(int en, void *target)
	{
		// Load extent from adequate file 

		// Evaluate correct file to use by substracting file
		// sizes until current address fits into a file
		UInt32 f=stripefunc(en,efd.stripeCount);
		UInt32 oa=address[en<<1];
		UInt32 bl=address[(en<<1)+1];

		// Check whether this is a fully zero extent
		if(bl==0xffffffff)
		{
			memset(target,0,extSize);
			return;
		}

		// Read extent from diskand transform it
		data[f].seek((Int64)oa*efd.blockSize);
		UInt32 tl=((bl+(efd.blockSize-1))/efd.blockSize)*efd.blockSize;

		UInt8 *trg=(UInt8*)target;

		if(efd.flags&ExtentFileDescriptor::COMPRESSED)
		{
			// The extent is compressed
			data[f].read(tempBuffer,tl);
			// Here we use zlib to decompress into the temp buffer 1
			DWORD dcl=extSize;
			uncompress(trg,&dcl,tempBuffer,bl);
		}
		else
		{
			// The extent is raw
			data[f].read(trg,tl);
		}
	}

	bool ExtentFile::getCube(int r, int x0, int y0, int z0, int x1, int y1, int z1, int c, void *target)
	{
		if(x1<x0 || y1<y0 || z1<z0 || r<0 || r>=zoomLevels || c<0 || c>=efd.channels)
			return false;

		// Get extents for top left and lower right corners
		int xo0=x0/realExtSizeX;
		int yo0=y0/realExtSizeY;
		int zo0=z0/realExtSizeZ;
		int xo1=x1/realExtSizeX;
		int yo1=y1/realExtSizeY;
		int zo1=z1/realExtSizeZ;
		int cs=efd.channelSize(efd.channelFormat[c]);
		int coff=0;
		for(int i=0;i<c;i++)
			coff+=efd.channelSize(efd.channelFormat[i]);
		coff*=extSizeX*extSizeY*extSizeZ;

		UInt8 *te=(UInt8*)malloc(extSize+efd.blockSize);
		UInt8 *tr=(UInt8*)target;

		// Walk through extents
		for(int k=zo0;k<=zo1;k++)
		{
			for(int j=yo0;j<=yo1;j++)
			{
				for(int i=xo0;i<=xo1;i++)
				{
					if(i>=0 && i<extX[r] && j>=0 && j<extY[r] && k>=0 && k<extZ[r])
					{
						int en=i+j*extX[r]+k*extX[r]*extY[r]+extOff[r];
						loadExtent(en,te);
					}
					else
						memset(te,0,extSize);

					// Walk throught voxels of extents
					for(int z=0;z<realExtSizeX;z++)
					{
						for(int y=0;y<realExtSizeY;y++)
						{
							for(int x=0;x<realExtSizeX;x++)
							{
								int fx=x+i*realExtSizeX;
								int fy=y+j*realExtSizeY;
								int fz=z+k*realExtSizeZ;
								if(fx>=x0 && fx<x1 && fy>=y0 && fy<y1 && fz>=z0 && fz<z1)
								{
									// If this voxel is in the destination, copy it out
									fx-=x0;
									fy-=y0;
									fz-=z0;
									for(int i=0;i<cs;i++)
										tr[(fx+fy*(x1-x0)+fz*(x1-x0)*(y1-y0))*cs+i]=te[(x+y*extSizeX+z*extSizeX*extSizeY)*cs+i+coff];
								}
							}
						}
					}
				}
			}
		}
		return true;
	}


	bool ExtentFile::open(const char *configFileName)
	{
		if(active)
			close();

		// Load XML configuration file
		efd.loadFromXML(configFileName);

		voxelSizeX=efd.voxelSize.x;
		voxelSizeY=efd.voxelSize.y;
		voxelSizeZ=efd.voxelSize.z;

		if(efd.labels)
			loadLabels(efd.labels);

		// Load address file
		char fname[1024], bname[1024];
		sprintf(fname,"%s/%s",(const char *)efd.stripes[0].path,(const char *)efd.stripes[0].index);
		FILE *faddr=fopen(fname,"rb");
		if(faddr==NULL)
			return false;
		fseek(faddr,0,2);
		int afs=ftell(faddr);
		fseek(faddr,0,0);
		address=(UInt32 *)malloc(afs);
		fread(address,afs,1,faddr);
		fclose(faddr);

		// Open stripes
		diskCount=0;
		for(int i=0;i<efd.stripeCount;i++)
		{
			sprintf(bname,"%s/%s",(const char *)efd.stripes[i].path,(const char *)efd.stripes[i].data);
			sprintf(fname,bname,0);
			if(!data[diskCount++].open(fname,FileWrapper::READ))
				return false;
		}

		// Copy out useful parameters from descriptor
		extSizeX=efd.fullExtentSize.x;
		extSizeY=efd.fullExtentSize.y;
		extSizeZ=efd.fullExtentSize.z;
		zoomLevels=efd.resolutions;
		extCount=efd.extOff[efd.resolutions];
		for(i=0;i<zoomLevels;i++)
		{
			extOff[i]=efd.extOff[i];
			extX[i]=efd.extX[i];
			extY[i]=efd.extY[i];
			extZ[i]=efd.extZ[i];
			sizeX[i]=efd.sizeX[i];
			sizeY[i]=efd.sizeY[i];
			sizeZ[i]=efd.sizeZ[i];
		}


		realExtSizeX=efd.extentSize.x;
		realExtSizeY=efd.extentSize.y;
		realExtSizeZ=efd.extentSize.z;
		imgSizeX=extX[0]*realExtSizeX;
		imgSizeY=extY[0]*realExtSizeY;
		imgSizeZ=extZ[0]*realExtSizeZ;

		// Allocate temporary buffer for decompression
		extSize=((efd.extSize+(efd.blockSize-1))/efd.blockSize)*efd.blockSize;
		tempBuffer=(UInt8 *)malloc(extSize);
		
		active=true;
		return true;
	}

	bool ExtentFile::close()
	{
		// Close all files and free all resources
		free(tempBuffer);

		free(address);
		for(int i=0;i<diskCount;i++)
			data[i].close();

		active=false;
		return true;
	}


}
